package ca.odell.glazedlists.matchers;

import javax.swing.event.EventListenerList;

/**
 * Basic building block for {@link MatcherEditor} implementations that
 * handles the details of dealing with listenerList. All {@link MatcherEditor}
 * implementators are encouraged to extends this class for its convenience methods.
 *
 * <p>Extending classes can fire events to listenerList using "fire" methods:
 * <ul>
 *    <li>{@link #fireMatchNone()}</li>
 *    <li>{@link #fireConstrained(Matcher)}</li>
 *    <li>{@link #fireChanged(Matcher)}</li>
 *    <li>{@link #fireRelaxed(Matcher)}</li>
 *    <li>{@link #fireMatchAll()}</li>
 * </ul>
 *
 * @author <a href="mailto:rob@starlight-systems.com">Rob Eden</a>
 */
public abstract class AbstractMatcherEditor implements MatcherEditor {
    
    public AbstractMatcherEditor() {
        super();
    }
    
    /**
     * listenerList for this Editor 
     */
    private EventListenerList listenerList = new EventListenerList();
    
    /**
     * the current Matcher in effect 
     */
    protected Matcher currentMatcher = Matchers.trueMatcher();
    
    /**
     * {@inheritDoc} 
     */
    public Matcher getMatcher() {
        return currentMatcher;
    }
    
    /**
     * {@inheritDoc} 
     */
    public final void addMatcherEditorListener(MatcherEditor.Listener listener) {
        listenerList.add(MatcherEditor.Listener.class, listener);
    }
    
    /**
     * {@inheritDoc} 
     */
    public final void removeMatcherEditorListener(MatcherEditor.Listener listener) {
        listenerList.remove(Listener.class, listener);
    }
    
    /**
     * Indicates that the filter matches all.
     */
    protected final void fireMatchAll() {
        this.currentMatcher = Matchers.trueMatcher();
        this.fireChangedMatcher(new MatcherEditor.Event(this, Event.MATCH_ALL, this.currentMatcher));
    }
    
    /**
     * Indicates that the filter has changed in an inditerminate way.
     */
    protected final void fireChanged(Matcher matcher) {
        if (matcher == null) throw new NullPointerException();
        this.currentMatcher = matcher;
        this.fireChangedMatcher(new MatcherEditor.Event(this, Event.CHANGED, this.currentMatcher));
    }
    
    /**
     * Indicates that the filter has changed to be more restrictive. This should only be
     * called if all currently filtered items will remain filtered.
     */
    protected final void fireConstrained(Matcher matcher) {
        if (matcher == null) throw new NullPointerException();
        this.currentMatcher = matcher;
        this.fireChangedMatcher(new MatcherEditor.Event(this, Event.CONSTRAINED, this.currentMatcher));
    }
    
    /**
     * Indicates that the filter has changed to be less restrictive. This should only be
     * called if all currently unfiltered items will remain unfiltered.
     */
    protected final void fireRelaxed(Matcher matcher) {
        if (matcher == null) throw new NullPointerException();
        this.currentMatcher = matcher;
        this.fireChangedMatcher(new MatcherEditor.Event(this, Event.RELAXED, this.currentMatcher));
    }
    
    /**
     * Indicates that the filter matches none.
     */
    protected final void fireMatchNone() {
        this.currentMatcher = Matchers.falseMatcher();
        this.fireChangedMatcher(new MatcherEditor.Event(this, Event.MATCH_NONE, this.currentMatcher));
    }
    
    protected final void fireChangedMatcher(MatcherEditor.Event event) {
        final Object[] listeners = this.listenerList.getListenerList();
        for (int i = listeners.length - 2; i >= 0; i -= 2) ((Listener)(MatcherEditor.Listener)listeners[i + 1]).changedMatcher(event);
    }
}
