package ca.odell.glazedlists;

import junit.framework.*;
import java.util.*;

/**
 * This test verifies that the ThresholdList works as expected.
 *
 * @author <a href="mailto:kevin@swank.ca">Kevin Maltby</a>
 */
public class ThresholdListTest extends TestCase {
    
    public ThresholdListTest() {
        super();
    }
    
    /**
     * the source list 
     */
    private BasicEventList source = null;
    
    /**
     * the threshold list to test with 
     */
    private ThresholdList thresholdList = null;
    
    /**
     * the enumerator to use to evaluate the objects in the list 
     */
    private IntegerEvaluator evaluator = null;
    
    /**
     * for randomly choosing list indicies 
     */
    private Random random = new Random(2);
    
    /**
     * Prepare for the test.
     */
    public void setUp() {
        source = new BasicEventList();
        evaluator = new IntegerEvaluator();
        thresholdList = new ThresholdList(source, evaluator);
        thresholdList.addListEventListener(new ListConsistencyListener(thresholdList, "thresholdList"));
    }
    
    /**
     * Clean up after the test.
     */
    public void tearDown() {
        thresholdList = null;
        source = null;
        random = null;
    }
    
    /**
     * Verifies that ThresholdList fires the right events.
     */
    public void testEventFiringLower() {
        source.addAll(Arrays.asList(new Integer[]{new Integer(25), new Integer(50), new Integer(50), new Integer(75), new Integer(75), new Integer(100)}));
        thresholdList.setLowerThreshold(27);
        thresholdList.setLowerThreshold(18);
    }
    
    /**
     * Verifies that ThresholdList fires the right events.
     */
    public void testEventFiringUpper() {
        source.addAll(Arrays.asList(new Integer[]{new Integer(25), new Integer(50), new Integer(50), new Integer(75), new Integer(75), new Integer(100)}));
        thresholdList.setUpperThreshold(77);
        thresholdList.setUpperThreshold(102);
    }
    
    /**
     * Verifies that ThresholdList fires the right events.
     */
    public void testEventFiringFromOneEntryToFullViaLower() {
        thresholdList.setUpperThreshold(100);
        thresholdList.setLowerThreshold(100);
        source.addAll(Arrays.asList(new Integer[]{new Integer(25), new Integer(50), new Integer(50), new Integer(75), new Integer(75), new Integer(100)}));
        assertEquals(1, thresholdList.size());
        thresholdList.setLowerThreshold(25);
        assertEquals(6, thresholdList.size());
    }
    
    /**
     * Verifies that ThresholdList fires the right events.
     */
    public void testEventFiringFromOneEntryToFullViaUpper() {
        thresholdList.setUpperThreshold(25);
        thresholdList.setLowerThreshold(25);
        source.addAll(Arrays.asList(new Integer[]{new Integer(25), new Integer(50), new Integer(50), new Integer(75), new Integer(75), new Integer(100)}));
        assertEquals(1, thresholdList.size());
        thresholdList.setUpperThreshold(100);
        assertEquals(6, thresholdList.size());
    }
    
    /**
     * A simple test to validate that ThresholdList is
     * behaving as expected on a list with no missing elements.
     */
    public void testListSimple() {
        ThresholdList lowerThresholdList = thresholdList;
        ThresholdList upperThresholdList = new ThresholdList(source, evaluator);
        for (int i = 0; i < 1000; i++) {
            source.add(i, new Integer(i));
        }
        for (int i = 0; i < 100; i++) {
            int num = random.nextInt(1000);
            lowerThresholdList.setLowerThreshold(num);
            upperThresholdList.setUpperThreshold(num);
            assertEquals(source.size() - num, lowerThresholdList.size());
            assertEquals(num + 1, upperThresholdList.size());
            assertEquals(new Integer(num), lowerThresholdList.get(0));
            assertEquals(new Integer(num), upperThresholdList.get(upperThresholdList.size() - 1));
        }
    }
    
    /**
     * A test to validate that ThresholdList is
     * behaving as expected on a list with missing elements.
     */
    public void testListWithMissingElements() {
        ThresholdList lowerThresholdList = thresholdList;
        ThresholdList upperThresholdList = new ThresholdList(source, evaluator);
        for (int i = 0; i < 1000; i++) {
            if (i % 5 == 0) source.add(i, new Integer(i - 1)); else source.add(i, new Integer(i));
        }
        lowerThresholdList.setLowerThreshold(2);
        upperThresholdList.setUpperThreshold(2);
        for (int i = 0; i < 1000; i += 5) {
            int num = i;
            lowerThresholdList.setLowerThreshold(num);
            upperThresholdList.setUpperThreshold(num);
            assertEquals(source.size() - num - 1, lowerThresholdList.size());
            assertEquals(num + 1, upperThresholdList.size());
            assertEquals(new Integer(num + 1), lowerThresholdList.get(0));
            assertEquals(new Integer(num - 1), upperThresholdList.get(upperThresholdList.size() - 1));
        }
    }
    
    /**
     * A test to validate that ThresholdList is behaving as expected
     * when values beyond the threshold are added for an increasing list.
     */
    public void testAddingValuesBeyondLowerThreshold() {
        source.add(0, new Integer(1));
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(0));
            source.add(i, new Integer(2));
        }
        thresholdList.setLowerThreshold(1);
        int originalSize = thresholdList.size();
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(0));
        }
        assertEquals(originalSize, thresholdList.size());
        for (int i = 0; i < thresholdList.size(); i++) {
            int enumeration = evaluator.evaluate(thresholdList.get(i));
            String message = "The value " + enumeration + " is beyond the threshold of 1.";
            assertEquals(message, true, (enumeration >= 1));
        }
    }
    
    /**
     * A test to validate that ThresholdList is behaving as expected
     * when values beyond the threshold are added for a decreasing list.
     */
    public void testAddingValuesBeyondUpperThreshold() {
        source.add(0, new Integer(1));
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(0));
            source.add(i, new Integer(2));
        }
        thresholdList.setUpperThreshold(1);
        int originalSize = thresholdList.size();
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(2));
        }
        assertEquals(originalSize, thresholdList.size());
        for (int i = 0; i < thresholdList.size(); i++) {
            String message = "The element at " + i + " is beyond the threshold.";
            assertEquals(message, true, (evaluator.evaluate(thresholdList.get(i)) <= 1));
        }
    }
    
    /**
     * A test to validate that ThresholdList is behaving as expected
     * when values within the threshold are added for an increasing list.
     */
    public void testAddingValuesWithinLowerThreshold() {
        source.add(0, new Integer(1));
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(0));
            source.add(i, new Integer(2));
        }
        thresholdList.setLowerThreshold(1);
        int originalSize = thresholdList.size();
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(2));
        }
        assertEquals(originalSize + 5, thresholdList.size());
        for (int i = 0; i < thresholdList.size(); i++) {
            int enumeration = evaluator.evaluate(thresholdList.get(i));
            String message = "The value " + enumeration + " is beyond the threshold of 1.";
            assertEquals(message, true, (enumeration >= 1));
        }
    }
    
    /**
     * A test to validate that ThresholdList is behaving as expected
     * when values within the threshold are added for a decreasing list.
     */
    public void testAddingValuesWithinUpperThreshold() {
        source.add(0, new Integer(1));
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(0));
            source.add(i, new Integer(2));
        }
        thresholdList.setUpperThreshold(1);
        int originalSize = thresholdList.size();
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(0));
        }
        assertEquals(originalSize + 5, thresholdList.size());
        for (int i = 0; i < thresholdList.size(); i++) {
            String message = "The element at " + i + " is beyond the threshold.";
            assertEquals(message, true, (evaluator.evaluate(thresholdList.get(i)) <= 1));
        }
    }
    
    /**
     * A test to validate that ThresholdList is behaving as expected
     * when values at the threshold are added for an increasing list.
     */
    public void testAddingValuesAtLowerThreshold() {
        source.add(0, new Integer(1));
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(0));
            source.add(i, new Integer(2));
        }
        thresholdList.setLowerThreshold(1);
        int originalSize = thresholdList.size();
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(1));
        }
        assertEquals(originalSize + 5, thresholdList.size());
        for (int i = 0; i < thresholdList.size(); i++) {
            int enumeration = evaluator.evaluate(thresholdList.get(i));
            String message = "The value " + enumeration + " is beyond the threshold of 1.";
            assertEquals(message, true, (enumeration >= 1));
        }
    }
    
    /**
     * A test to validate that ThresholdList is behaving as expected
     * when values at the threshold are added for a decreasing list.
     */
    public void testAddingValuesAtUpperThreshold() {
        source.add(0, new Integer(1));
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(0));
            source.add(i, new Integer(2));
        }
        thresholdList.setUpperThreshold(1);
        int originalSize = thresholdList.size();
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(1));
        }
        assertEquals(originalSize + 5, thresholdList.size());
        for (int i = 0; i < thresholdList.size(); i++) {
            String message = "The element at " + i + " is beyond the threshold.";
            assertEquals(message, true, (evaluator.evaluate(thresholdList.get(i)) <= 1));
        }
    }
    
    /**
     * Test an edge case where no value is at the current threshold
     * and then a value is added which is at the threshold.
     */
    public void testAddingAtLowerThresholdWithNoValuesAtThreshold() {
        source.add(0, new Integer(0));
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(5));
            source.add(i, new Integer(10));
        }
        thresholdList.setLowerThreshold(3);
        int originalSize = thresholdList.size();
        source.add(0, new Integer(4));
        assertEquals(originalSize + 1, thresholdList.size());
        assertEquals(new Integer(4), thresholdList.get(0));
        originalSize = thresholdList.size();
        source.add(0, new Integer(3));
        assertEquals(originalSize + 1, thresholdList.size());
        assertEquals(new Integer(3), thresholdList.get(0));
    }
    
    /**
     * Test an edge case where no value is at the current threshold
     * and then a value is added which is at the threshold.
     */
    public void testAddingAtUpperThresholdWithNoValuesAtThreshold() {
        source.add(0, new Integer(5));
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(0));
            source.add(i, new Integer(10));
        }
        thresholdList.setUpperThreshold(7);
        int originalSize = thresholdList.size();
        source.add(0, new Integer(6));
        assertEquals(originalSize + 1, thresholdList.size());
        assertEquals(new Integer(6), thresholdList.get(thresholdList.size() - 1));
        originalSize = thresholdList.size();
        source.add(0, new Integer(7));
        assertEquals(originalSize + 1, thresholdList.size());
        assertEquals(new Integer(7), thresholdList.get(thresholdList.size() - 1));
    }
    
    /**
     * A test to validate that ThresholdList is behaving as expected
     * when values beyond the threshold are removed for an increasing list.
     */
    public void testRemovingValuesBeyondLowerThreshold() {
        source.add(0, new Integer(1));
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(0));
            source.add(i, new Integer(2));
        }
        thresholdList.setLowerThreshold(1);
        int originalSize = thresholdList.size();
        for (int i = 0; i < 4; i++) {
            source.add(i, new Integer(0));
        }
        assertEquals(originalSize, thresholdList.size());
        for (int i = 0; i < thresholdList.size(); i++) {
            int enumeration = evaluator.evaluate(thresholdList.get(i));
            String message = "The value " + enumeration + " is beyond the threshold of 1.";
            assertEquals(message, true, (enumeration >= 1));
        }
    }
    
    /**
     * A test to validate that ThresholdList is behaving as expected
     * when values beyond the threshold are removed for a decreasing list.
     */
    public void testRemovingValuesBeyondUpperThreshold() {
        source.add(0, new Integer(1));
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(0));
            source.add(i, new Integer(2));
        }
        thresholdList.setUpperThreshold(1);
        int originalSize = thresholdList.size();
        for (int i = 0; i < 4; i++) {
            source.remove(new Integer(2));
        }
        assertEquals(originalSize, thresholdList.size());
        for (int i = 0; i < thresholdList.size(); i++) {
            String message = "The element at " + i + " is beyond the threshold.";
            assertEquals(message, true, (evaluator.evaluate(thresholdList.get(i)) <= 1));
        }
    }
    
    /**
     * A test to validate that ThresholdList is behaving as expected
     * when values within the threshold are removed for an increasing list.
     */
    public void testRemovingValuesWithinLowerThreshold() {
        source.add(0, new Integer(1));
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(0));
            source.add(i, new Integer(2));
        }
        thresholdList.setLowerThreshold(1);
        int originalSize = thresholdList.size();
        for (int i = 0; i < 4; i++) {
            source.remove(new Integer(2));
        }
        assertEquals(originalSize - 4, thresholdList.size());
        for (int i = 0; i < thresholdList.size(); i++) {
            int enumeration = evaluator.evaluate(thresholdList.get(i));
            String message = "The value " + enumeration + " is beyond the threshold of 1.";
            assertEquals(message, true, (enumeration >= 1));
        }
    }
    
    /**
     * A test to validate that ThresholdList is behaving as expected
     * when values within the threshold are removed for a decreasing list.
     */
    public void testRemovingValuesWithinUpperThreshold() {
        source.add(0, new Integer(1));
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(0));
            source.add(i, new Integer(2));
        }
        thresholdList.setUpperThreshold(1);
        int originalSize = thresholdList.size();
        for (int i = 0; i < 4; i++) {
            source.remove(new Integer(0));
        }
        assertEquals(originalSize - 4, thresholdList.size());
        for (int i = 0; i < thresholdList.size(); i++) {
            String message = "The element at " + i + " is beyond the threshold.";
            assertEquals(message, true, (evaluator.evaluate(thresholdList.get(i)) <= 1));
        }
    }
    
    /**
     * A test to validate that ThresholdList is behaving as expected
     * when values at the threshold are removed for an increasing list.
     */
    public void testRemovingValuesAtLowerThreshold() {
        source.add(0, new Integer(1));
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(0));
            source.add(i, new Integer(2));
        }
        thresholdList.setLowerThreshold(1);
        int originalSize = thresholdList.size();
        source.remove(new Integer(1));
        assertEquals(originalSize - 1, thresholdList.size());
        for (int i = 0; i < thresholdList.size(); i++) {
            int enumeration = evaluator.evaluate(thresholdList.get(i));
            String message = "The value " + enumeration + " is beyond the threshold of 1.";
            assertEquals(message, true, (enumeration >= 1));
        }
    }
    
    /**
     * A test to validate that ThresholdList is behaving as expected
     * when values at the threshold are removed for a decreasing list.
     */
    public void testRemovingValuesAtUpperThreshold() {
        source.add(0, new Integer(1));
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(0));
            source.add(i, new Integer(2));
        }
        thresholdList.setUpperThreshold(1);
        int originalSize = thresholdList.size();
        source.remove(new Integer(1));
        assertEquals(originalSize - 1, thresholdList.size());
        for (int i = 0; i < thresholdList.size(); i++) {
            String message = "The element at " + i + " is beyond the threshold.";
            assertEquals(message, true, (evaluator.evaluate(thresholdList.get(i)) <= 1));
        }
    }
    
    /**
     * A test to validate that ThresholdList is behaving as expected
     * when values beyond the threshold are updated for an increasing list.
     */
    public void testUpdatingValuesBeyondLowerThreshold() {
        source.add(0, new Integer(1));
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(2));
            source.add(0, new Integer(0));
        }
        thresholdList.setLowerThreshold(1);
        int originalSize = thresholdList.size();
        for (int i = 0; i < 5; i++) {
            source.set(i, new Integer(-1 - i));
        }
        assertEquals(originalSize, thresholdList.size());
        for (int i = 0; i < thresholdList.size(); i++) {
            int enumeration = evaluator.evaluate(thresholdList.get(i));
            String message = "The value " + enumeration + " is beyond the threshold of 1.";
            assertEquals(message, true, (enumeration >= 1));
        }
    }
    
    /**
     * A test to validate that ThresholdList is behaving as expected
     * when values beyond the threshold are updated for a decreasing list.
     */
    public void testUpdatingValuesBeyondUpperThreshold() {
        source.add(0, new Integer(1));
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(0));
            source.add(0, new Integer(2));
        }
        thresholdList.setUpperThreshold(1);
        int originalSize = thresholdList.size();
        for (int i = 0; i < 5; i++) {
            source.set(i, new Integer(3));
        }
        assertEquals(originalSize, thresholdList.size());
        for (int i = 0; i < thresholdList.size(); i++) {
            String message = "The element at " + i + " is beyond the threshold.";
            assertEquals(message, true, (evaluator.evaluate(thresholdList.get(i)) <= 1));
        }
    }
    
    /**
     * A test to validate that ThresholdList is behaving as expected
     * when values within the threshold are updated for an increasing list.
     */
    public void testUpdatingValuesWithinLowerThreshold() {
        source.add(0, new Integer(1));
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(0));
            source.add(0, new Integer(2));
        }
        thresholdList.setLowerThreshold(1);
        int originalSize = thresholdList.size();
        for (int i = 0; i < 5; i++) {
            source.set(i, new Integer(i + 3));
        }
        assertEquals(originalSize, thresholdList.size());
        for (int i = 0; i < thresholdList.size(); i++) {
            int enumeration = evaluator.evaluate(thresholdList.get(i));
            String message = "The value " + enumeration + " is beyond the threshold of 1.";
            assertEquals(message, true, (enumeration >= 1));
        }
    }
    
    /**
     * A test to validate that ThresholdList is behaving as expected
     * when values within the threshold are updated for a decreasing list.
     */
    public void testUpdatingValuesWithinUpperThreshold() {
        source.add(0, new Integer(1));
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(2));
            source.add(0, new Integer(0));
        }
        thresholdList.setUpperThreshold(1);
        int originalSize = thresholdList.size();
        for (int i = 0; i < 5; i++) {
            source.set(i, new Integer(-1 - i));
        }
        assertEquals(originalSize, thresholdList.size());
        for (int i = 0; i < thresholdList.size(); i++) {
            String message = "The element at " + i + " is beyond the threshold.";
            assertEquals(message, true, (evaluator.evaluate(thresholdList.get(i)) <= 1));
        }
    }
    
    /**
     * A test to validate that ThresholdList is behaving as expected
     * when values at the threshold are updated for an increasing list.
     */
    public void testUpdatingValuesAtLowerThreshold() {
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(0));
            source.add(i, new Integer(10));
        }
        source.add(0, new Integer(5));
        thresholdList.setLowerThreshold(5);
        int originalSize = thresholdList.size();
        source.set(0, new Integer(5));
        assertEquals(originalSize, thresholdList.size());
        assertEquals(new Integer(5), thresholdList.get(0));
        source.set(0, new Integer(6));
        assertEquals(originalSize, thresholdList.size());
        assertEquals(new Integer(6), thresholdList.get(0));
        source.set(0, new Integer(5));
        source.set(0, new Integer(4));
        assertEquals(originalSize - 1, thresholdList.size());
        assertEquals(new Integer(10), thresholdList.get(0));
        source.set(0, new Integer(5));
        assertEquals(originalSize, thresholdList.size());
        assertEquals(new Integer(5), thresholdList.get(0));
    }
    
    /**
     * A test to validate that ThresholdList is behaving as expected
     * when values at the threshold are updated for a decreasing list.
     */
    public void testUpdatingValuesAtUpperThreshold() {
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(0));
            source.add(i, new Integer(10));
        }
        source.add(0, new Integer(5));
        thresholdList.setUpperThreshold(5);
        int originalSize = thresholdList.size();
        source.set(0, new Integer(5));
        assertEquals(originalSize, thresholdList.size());
        assertEquals(new Integer(5), thresholdList.get(thresholdList.size() - 1));
        source.set(0, new Integer(4));
        assertEquals(originalSize, thresholdList.size());
        assertEquals(new Integer(4), thresholdList.get(thresholdList.size() - 1));
        source.set(0, new Integer(5));
        source.set(0, new Integer(6));
        assertEquals(originalSize - 1, thresholdList.size());
        assertEquals(new Integer(0), thresholdList.get(thresholdList.size() - 1));
        source.set(0, new Integer(5));
        assertEquals(originalSize, thresholdList.size());
        assertEquals(new Integer(5), thresholdList.get(thresholdList.size() - 1));
    }
    
    /**
     * Test an edge case where no value is at the current threshold
     * and then a value is added which is at the threshold.
     */
    public void testUpdatingAtLowerThresholdWithNoValuesAtThreshold() {
        source.add(0, new Integer(5));
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(0));
            source.add(i, new Integer(10));
        }
        thresholdList.setLowerThreshold(3);
        int originalSize = thresholdList.size();
        source.add(0, new Integer(2));
        int secondSize = thresholdList.size();
        assertEquals(originalSize, secondSize);
        source.set(0, new Integer(4));
        assertEquals(originalSize + 1, thresholdList.size());
        assertEquals(new Integer(4), thresholdList.get(0));
        originalSize = thresholdList.size();
        source.set(0, new Integer(1));
        assertEquals(originalSize - 1, thresholdList.size());
        assertEquals(new Integer(5), thresholdList.get(0));
        originalSize = thresholdList.size();
        source.set(0, new Integer(3));
        assertEquals(originalSize + 1, thresholdList.size());
        assertEquals(new Integer(3), thresholdList.get(0));
    }
    
    /**
     * Test an edge case where no value is at the current threshold
     * and then a value is added which is at the threshold.
     */
    public void testUpdatingAtUpperThresholdWithNoValuesAtThreshold() {
        source.add(0, new Integer(5));
        for (int i = 0; i < 5; i++) {
            source.add(i, new Integer(0));
            source.add(i, new Integer(10));
        }
        thresholdList.setUpperThreshold(7);
        int originalSize = thresholdList.size();
        source.add(0, new Integer(8));
        source.set(0, new Integer(6));
        assertEquals(originalSize + 1, thresholdList.size());
        assertEquals(new Integer(6), thresholdList.get(thresholdList.size() - 1));
        originalSize = thresholdList.size();
        source.set(0, new Integer(9));
        assertEquals(originalSize - 1, thresholdList.size());
        assertEquals(new Integer(5), thresholdList.get(thresholdList.size() - 1));
        originalSize = thresholdList.size();
        source.set(0, new Integer(7));
        assertEquals(originalSize + 1, thresholdList.size());
        assertEquals(new Integer(7), thresholdList.get(thresholdList.size() - 1));
    }
    
    /**
     * A test to validate the size of an empty size and that the
     * size is still zero once a list is filled and emptied.
     */
    public void testSizeZeroing() {
        ThresholdList lowerThresholdList = thresholdList;
        ThresholdList upperThresholdList = new ThresholdList(source, evaluator);
        assertEquals(0, lowerThresholdList.size());
        assertEquals(0, upperThresholdList.size());
        for (int i = 0; i < 10; i++) {
            source.add(i, new Integer(i));
        }
        lowerThresholdList.setLowerThreshold(11);
        upperThresholdList.setUpperThreshold(-1);
        assertEquals(0, lowerThresholdList.size());
        assertEquals(0, upperThresholdList.size());
        for (int i = 0; i < 10; i++) {
            source.remove(0);
        }
        assertEquals(0, lowerThresholdList.size());
        assertEquals(0, upperThresholdList.size());
    }
    
    /**
     * A test to validate events from calls to the threshold setting methods
     */
    public void testThresholdSettingEvents() {
        ThresholdList lowerThresholdList = thresholdList;
        ThresholdList upperThresholdList = new ThresholdList(source, evaluator);
        for (int i = 0; i < 1000; i++) {
            source.add(i, new Integer(i));
        }
        SortedList sortedIncreasing = new SortedList(lowerThresholdList, new ThresholdComparator(new IntegerEvaluator()));
        SortedList sortedDecreasing = new SortedList(upperThresholdList, new ThresholdComparator(new IntegerEvaluator()));
        validateListsEquals(lowerThresholdList, sortedIncreasing);
        validateListsEquals(upperThresholdList, sortedDecreasing);
        for (int i = 0; i < 100; i++) {
            int num = random.nextInt(1000);
            lowerThresholdList.setLowerThreshold(num);
            upperThresholdList.setUpperThreshold(num);
            validateListsEquals(lowerThresholdList, sortedIncreasing);
            validateListsEquals(upperThresholdList, sortedDecreasing);
        }
    }
    
    /**
     * Test the edge conditions where a source list contains values and the
     * thresholds are set such that the list is empty.
     */
    public void testZeroSizeViewOfArbitraryList() {
        thresholdList.setLowerThreshold(-5);
        thresholdList.setUpperThreshold(-3);
        for (int i = 0; i < 1000; i++) {
            source.add(i, new Integer(-2 + random.nextInt(1004)));
        }
        assertEquals(0, thresholdList.size());
    }
    
    /**
     * Tests to see if pre-built source lists are initialized correctly on a
     * freshly constructed ThresholdList without setting any thresholds.
     */
    public void testInitialization() {
        SortedList sortedBase = new SortedList(source, new ThresholdComparator(new IntegerEvaluator()));
        for (int i = 0; i < 500; i++) {
            sortedBase.add(i, new Integer(i));
        }
        thresholdList = new ThresholdList(sortedBase, new IntegerEvaluator());
        SortedList sortedCover = new SortedList(thresholdList, new ThresholdComparator(new IntegerEvaluator()));
        validateListsEquals(sortedBase, sortedCover);
    }
    
    /**
     * Tests to see if events are handled correctly by a freshly
     * constructed ThresholdList without setting any thresholds.
     */
    public void testEventsOnNewThresholdList() {
        SortedList sortedBase = new SortedList(source, new ThresholdComparator(new IntegerEvaluator()));
        thresholdList = new ThresholdList(sortedBase, new IntegerEvaluator());
        SortedList sortedCover = new SortedList(thresholdList, new ThresholdComparator(new IntegerEvaluator()));
        for (int i = 0; i < 500; i++) {
            source.add(i, new Integer(i));
        }
        validateListsEquals(sortedBase, sortedCover);
    }
    
    /**
     * Verifies that the ThresholdList fires consistent events.
     */
    public void testEventFiringIsConsistent() {
        thresholdList.setLowerThreshold(-14931);
        thresholdList.setUpperThreshold(-1931);
        thresholdList.setLowerThreshold(-13931);
        thresholdList.setUpperThreshold(-931);
        source.add(new Integer(1401));
        thresholdList.setLowerThreshold(-12926);
        thresholdList.setUpperThreshold(74);
        source.add(new Integer(2168));
        source.add(new Integer(2996));
        thresholdList.setLowerThreshold(-11930);
        thresholdList.setUpperThreshold(1070);
        source.add(new Integer(3895));
        thresholdList.setLowerThreshold(-10931);
        thresholdList.setUpperThreshold(2069);
        source.add(new Integer(4235));
        source.add(new Integer(4507));
        thresholdList.setLowerThreshold(-9931);
        thresholdList.setUpperThreshold(3069);
    }
    
    /**
     * Verifies that the ThresholdList fires consistent events.
     */
    public void testNumberOfEventsFired() {
        ListEventCounter counter = new ListEventCounter();
        thresholdList.addListEventListener(counter);
        thresholdList.setLowerThreshold(-14922);
        thresholdList.setUpperThreshold(-1922);
        thresholdList.setLowerThreshold(-13923);
        thresholdList.setUpperThreshold(-923);
        assertEquals(0, counter.getEventCount());
        source.add(new Integer(1139));
        assertEquals(0, counter.getEventCount());
        thresholdList.setLowerThreshold(-12923);
        thresholdList.setUpperThreshold(77);
        thresholdList.setLowerThreshold(-11922);
        thresholdList.setUpperThreshold(1078);
        thresholdList.setLowerThreshold(-10923);
        thresholdList.setUpperThreshold(2077);
        assertEquals(source, thresholdList);
        assertEquals(1, counter.getEventCount());
        assertEquals(1, counter.getChangeCount(0));
        thresholdList.setLowerThreshold(-9923);
        thresholdList.setUpperThreshold(3077);
        thresholdList.setLowerThreshold(-8923);
        thresholdList.setUpperThreshold(4077);
        thresholdList.setLowerThreshold(-7921);
        thresholdList.setUpperThreshold(5079);
        thresholdList.setLowerThreshold(-6923);
        thresholdList.setUpperThreshold(6077);
        thresholdList.setLowerThreshold(-5923);
        thresholdList.setUpperThreshold(7077);
        thresholdList.setLowerThreshold(-4922);
        thresholdList.setUpperThreshold(8078);
        thresholdList.setLowerThreshold(-3923);
        thresholdList.setUpperThreshold(9077);
        thresholdList.setLowerThreshold(-2923);
        thresholdList.setUpperThreshold(10077);
        thresholdList.setLowerThreshold(-1922);
        thresholdList.setUpperThreshold(11078);
        thresholdList.setLowerThreshold(-923);
        thresholdList.setUpperThreshold(12077);
        thresholdList.setLowerThreshold(77);
        thresholdList.setUpperThreshold(13077);
        thresholdList.setLowerThreshold(1077);
        thresholdList.setUpperThreshold(14077);
        thresholdList.setLowerThreshold(2078);
        assertEquals(2, counter.getEventCount());
        assertEquals(1, counter.getChangeCount(1));
        assertEquals(Collections.EMPTY_LIST, thresholdList);
    }
    
    /**
     * Tests that a increasing sliding windows works. This is the case where
     * the upper and lower thresholds keep keep shifting to higher values.
     */
    public void testIncreasingSlidingWindow() {
        ListEventCounter counter = new ListEventCounter();
        thresholdList.addListEventListener(counter);
        thresholdList.setLowerThreshold(-14938);
        thresholdList.setUpperThreshold(-1938);
        source.add(new Integer(995));
        thresholdList.setLowerThreshold(-13937);
        thresholdList.setUpperThreshold(-937);
        source.add(new Integer(1155));
        thresholdList.setLowerThreshold(4062);
        thresholdList.setUpperThreshold(17062);
        source.add(new Integer(19970));
        thresholdList.setLowerThreshold(5063);
        thresholdList.setUpperThreshold(18063);
        source.add(new Integer(20279));
        thresholdList.setLowerThreshold(7063);
        thresholdList.setUpperThreshold(20063);
    }
    
    /**
     * Tests that a decreasing sliding windows works. This is the case where
     * both the upper and lower thresholds keep shifting to lower values.
     */
    public void testDecreasingSlidingWindow() {
        ListEventCounter counter = new ListEventCounter();
        thresholdList.addListEventListener(counter);
        thresholdList.setUpperThreshold(20063);
        thresholdList.setLowerThreshold(7063);
        source.add(new Integer(7000));
        thresholdList.setUpperThreshold(6000);
        thresholdList.setLowerThreshold(0);
    }
    
    /**
     * Tests that the JavaBean constructor and supporting code
     * works as expected.
     */
    public void testJavaBeanConstructor() {
        thresholdList.dispose();
        thresholdList = null;
        thresholdList = new ThresholdList(source, "value");
        for (int i = 0; i < 1000; i++) {
            source.add(new SimpleJavaBeanObject(i));
        }
        thresholdList.setLowerThreshold(500);
        assertEquals(500, thresholdList.size());
    }
    
    public void testBoundaryConditions() {
        source.add(new Integer(0));
        source.add(new Integer(25));
        source.add(new Integer(50));
        source.add(new Integer(75));
        source.add(new Integer(100));
        thresholdList.setLowerThreshold(0);
        thresholdList.setUpperThreshold(100);
        assertEquals(5, thresholdList.size());
        thresholdList.setLowerThreshold(100);
        assertEquals(1, thresholdList.size());
    }
    
    public void testNonIntegers() {
        thresholdList.dispose();
        thresholdList = new ThresholdList(source, new IntegerAsStringEvaluator());
        source.add("0");
        source.add("25");
        source.add("50");
        source.add("75");
        source.add("100");
        thresholdList.setLowerThreshold(0);
        thresholdList.setUpperThreshold(100);
        assertEquals(5, thresholdList.size());
        thresholdList.setLowerThreshold(100);
        assertEquals(1, thresholdList.size());
    }
    
    /**
     * A helper method to compare two lists for equality
     */
    private void validateListsEquals(EventList alpha, EventList beta) {
        assertEquals(alpha.size(), beta.size());
        for (int i = 0; i < alpha.size(); i++) {
            assertEquals(alpha.get(i), beta.get(i));
        }
    }
    
    /**
     * A JavaBean to test out the Reflection constructor
     */
    public class SimpleJavaBeanObject {
        private int value = 0;
        
        public SimpleJavaBeanObject(int value) {
            super();
            this.value = value;
        }
        
        public int getValue() {
            return value;
        }
    }
    
    /**
     * Returns an integer value which represents the object.
     */
    private class IntegerEvaluator implements ThresholdList.Evaluator {
        
        private IntegerEvaluator() {
            super();
        }
        
        public int evaluate(Object object) {
            if (object == null) {
                return Integer.MIN_VALUE;
            } else if (!(object instanceof Integer)) {
                throw new IllegalArgumentException("The value passed to this enumerator is of an invalid type. All elements in the list must be of type Integer.");
            } else {
                Integer integer = (Integer)(Integer)object;
                return integer.intValue();
            }
        }
    }
    
    /**
     * Expects strings like "100" and "50" and returns their integer values.
     */
    private class IntegerAsStringEvaluator implements ThresholdList.Evaluator {
        
        private IntegerAsStringEvaluator() {
            super();
        }
        
        public int evaluate(Object object) {
            return Integer.valueOf((String)(String)object);
        }
    }
    
    /**
     * A ThresholdComparator is a simple helper class that wraps
     * a <code>ThresholdEvaluator</code> with a <code>Comparator</code> to
     * be used for sorting of the <code>ThresholdList</code>.
     *
     * <p>This class is duplicated within this testcase for the
     * testThresholdSettingEvents test.  The code duplication exists only
     * for convienience and for the sake of that one test.
     */
    private final class ThresholdComparator implements Comparator {
        
        /**
         * the underlying evaluator 
         */
        private ThresholdList.Evaluator evaluator = null;
        
        /**
         * Creates a new ThresholdComparator
         */
        ThresholdComparator(ThresholdList.Evaluator evaluator) {
            super();
            this.evaluator = evaluator;
        }
        
        /**
         * Compares two <code>Object</code>s, and compares them using the result
         * given when each <code>Object</code> is evaluated using the underlying
         * <code>ThresholdEvaluator</code>.
         */
        public int compare(Object alpha, Object beta) {
            int alphaValue = 0;
            if (alpha instanceof Integer) alphaValue = ((Integer)(Integer)alpha).intValue(); else alphaValue = evaluator.evaluate(alpha);
            int betaValue = 0;
            if (beta instanceof Integer) betaValue = ((Integer)(Integer)beta).intValue(); else betaValue = evaluator.evaluate(beta);
            if (alphaValue > betaValue) return 1; else if (alphaValue < betaValue) return -1; else return 0;
        }
        
        /**
         * Returns true iff the object passed is a <code>ThresholdComparator</code> with
         * the same underlying <code>ThresholdEvaluator</code>.
         */
        public boolean equals(Object object) {
            if (object == null || !(object instanceof ThresholdComparator)) {
                return false;
            }
            ThresholdComparator other = (ThresholdComparator)(ThresholdListTest.ThresholdComparator)object;
            return this.evaluator == other.evaluator;
        }
    }
}
